package com.jeffknecht.dynadao.api;

import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.Proxy;
import java.util.HashMap;
import java.util.List;

import javax.sql.DataSource;

import net.sf.cglib.core.MethodInfo;
import net.sf.cglib.proxy.Callback;
import net.sf.cglib.proxy.Enhancer;
import net.sf.cglib.proxy.MethodInterceptor;
import net.sf.cglib.proxy.MethodProxy;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.jeffknecht.dynadao.annotation.DTOClass;
import com.jeffknecht.dynadao.annotation.GroupEnd;
import com.jeffknecht.dynadao.annotation.GroupStart;
import com.jeffknecht.dynadao.annotation.Or;
import com.jeffknecht.dynadao.annotation.OrderBy;
import com.jeffknecht.dynadao.annotation.Parameter;
import com.jeffknecht.dynadao.annotation.Parameter.Operator;
import com.jeffknecht.dynadao.annotation.Where;
import com.jeffknecht.dynadao.impl.DynaDaoImpl;

/**
 * <p>
 * Provides a dynamic proxy for a DAO interface.
 * </p>
 * 
 * <p>
 * There are some specific rules the DAO interface must follow:
 * </p>
 * 
 * <p>First, the interface may only declare the following methods:
 * <ul>
 * <li>int insert(Object<T>)</li>
 * <li>int update(Object<T>)</li>
 * <li>int delete(Object<T>)</li>
 * <li>int countAll()</li>
 * <li>int countBy...(args...)</li>
 * <li>List<T> findAll()</li>
 * <li>T findBy...(args...)</li>
 * <li>List<T> findBy...(args...)</li>
 * </ul>
 * </p>
 * 
 * <p>Second, either the interface or each of the find methods
 * must be annotated with the DTOClass annotation, defining
 * the class (<T>) that is returned by the find... methods</p>
 * 
 * <p>Third, every parameter of a find... method must be
 * annotated with a Parameter annotation, defining the property
 * in the DTOClass that is represented by the parameter</p>
 * 
 * <p>Any call to a method that does not follow these guidelines
 * may be met with a nasty runtime exception. You have been warned!</p>
 * 
 * @author jeff.knecht
 *
 */
public class DynaDaoProxy {

	/**
	 * Builds and returns a DynaDao proxy based on the supplied interface.
	 * 
	 * @param intfce The interface to proxy
	 * @param dynaDao The DynaDao instance to pass the calls to.
	 * @return
	 */
	public static <T> T create(Class<T> clazz, DynaDao dynaDao) {
		return (T) Enhancer.create(clazz, new DynaDaoProxyCallback(clazz, dynaDao));
	}
	
	/**
	 * Builds and returns a DynaDao proxy based on the supplied interface.
	 * 
	 * @param intfce The interface to proxy
	 * @param dataSource The dataSource to use for connections to the database.
	 * @return
	 */
	public static <T> T create(Class<T> clazz, DataSource dataSource) {
		DynaDao dynaDao = new DynaDaoImpl(dataSource);
		return (T) Enhancer.create(clazz, new DynaDaoProxyCallback(clazz, dynaDao));
	}
	
	private static final class DynaDaoProxyCallback implements MethodInterceptor {
		private static final Log log = LogFactory.getLog(DynaDaoProxy.class);
		private DynaDao dynaDao;
		private Class<?> proxyClazz;
		private Log proxyLog;

		private DynaDaoProxyCallback(Class proxyClazz, DynaDao dynaDao) {
			this.dynaDao = dynaDao;
			this.proxyClazz = proxyClazz;
			this.proxyLog = LogFactory.getLog(proxyClazz);
		}

		public Object intercept(Object obj, Method method, Object[] args,
				MethodProxy proxy) throws Throwable {
			if (!Modifier.isAbstract(method.getModifiers())) {
				return proxy.invokeSuper(obj, args);
			} else {
				String methodName = method.getName();
				
				if (log.isDebugEnabled()) {
					log.debug("proxying method " + methodName + " for class " + proxyClazz);
				}
				
				if (methodName.startsWith("insert")) {
					if (proxyLog.isDebugEnabled()) {
						proxyLog.debug(methodName + " called");
					}
					return dynaDao.insert(args[0]);
				} else if (methodName.startsWith("update")) {
					if (proxyLog.isDebugEnabled()) {
						proxyLog.debug(methodName + " called");
					}
					return dynaDao.update(args[0]);
				} else if (methodName.startsWith("delete")) {
					if (proxyLog.isDebugEnabled()) {
						proxyLog.debug(methodName + " called");
					}
					return dynaDao.delete(args[0]);
				} else if (methodName.equals("countAll")) {
					if (proxyLog.isDebugEnabled()) {
						proxyLog.debug(methodName + " called");
					}
					DTOClass type = getDTOClass(method);
					return dynaDao.countAll(type.value());
				} else if (methodName.startsWith("count")) {
					if (proxyLog.isDebugEnabled()) {
						proxyLog.debug(methodName + " called");
					}
					DTOClass type = getDTOClass(method);
					
					HashMap<String, Object> params = new HashMap<String, Object>();
					StringBuilder sb = new StringBuilder();
					
					buildDynaDaoWhereClause(method, args, params, sb);
				
					return dynaDao.count(type.value(), sb.toString(), params);
				} else if (methodName.equals("findAll")) {
					if (proxyLog.isDebugEnabled()) {
						proxyLog.debug(methodName + " called");
					}
					DTOClass type = getDTOClass(method);
					OrderBy orderBy = method.getAnnotation(OrderBy.class);
					if (orderBy != null) {
						return dynaDao.findAll(type.value(), orderBy.value());					
					} else {
						return dynaDao.findAll(type.value());
					}
				} else if (methodName.startsWith("find")) {
					if (proxyLog.isDebugEnabled()) {
						proxyLog.debug(methodName + " called");
					}
					DTOClass type = getDTOClass(method);
					
					HashMap<String, Object> params = new HashMap<String, Object>();
					StringBuilder sb = new StringBuilder();
					
					buildDynaDaoWhereClause(method, args, params, sb);
				
					if (method.getReturnType().isAssignableFrom(List.class)) {
						OrderBy orderBy = method.getAnnotation(OrderBy.class);
						if (orderBy != null) {
							return dynaDao.find(type.value(), sb.toString(), params, orderBy.value());
						} else {
							return dynaDao.find(type.value(), sb.toString(), params);
						}
					} else {
						return dynaDao.findOne(type.value(), sb.toString(), params);
					}
				} else {
					throw new UnsupportedOperationException("Unsupported method: " + methodName);
				}				
			}
		}

		
		/**
		 * 
		 * @param method The method that is being called
		 * @param args The argument values for the method
		 * @param params A Map to hold the generated parameter values
		 * @param sb A StringBuilder to append the where clause to
		 */
		private void buildDynaDaoWhereClause(Method method, Object[] args,
				HashMap<String, Object> params, StringBuilder sb) {
			
			StringBuilder staticWhereBuilder = new StringBuilder();
			Where whereAnnotation = method.getAnnotation(Where.class);
			if (whereAnnotation != null) {
				staticWhereBuilder.append(whereAnnotation.value().trim());
			}
			
			StringBuilder dynamicWhereBuilder = new StringBuilder();
			Annotation[][] paramAnnotations = method.getParameterAnnotations();
			
			int openParens = 0;
			for (int i = 0; i < paramAnnotations.length; i++) {
				boolean newParam = true;
				for (int j = 0; j < paramAnnotations[i].length; j++) {
					Annotation paramAnnot = paramAnnotations[i][j];
					if (newParam) {
						if (paramAnnot instanceof Or) {
							if (dynamicWhereBuilder.length() > 0) {
								dynamicWhereBuilder.append(" or ");
							}
						} else {
							if (dynamicWhereBuilder.length() > 0) {
								dynamicWhereBuilder.append(" and ");
							}
						}
						newParam = false;
					}
					

					if (paramAnnot instanceof GroupStart) {
						openParens++;
						dynamicWhereBuilder.append("(");
					}
					
					if (paramAnnot instanceof GroupEnd) {
						openParens--;
						dynamicWhereBuilder.append(")");
					}
					
					if (paramAnnot instanceof Parameter) {
						String paramName = ((Parameter) paramAnnot).value();
						String placeHolder = ((Parameter) paramAnnot).placeholder();
						
						//it can only really be case-insensitive if the parameter data type is string-ish
						boolean caseInsensitive = ((Parameter) paramAnnot).caseInsensitive() && (args[i] instanceof CharSequence || args[i] instanceof Character || args[i].getClass().equals(char.class));
						Operator comparison = ((Parameter) paramAnnot).comparison();
						boolean isComparisonIn = comparison.equals(Operator.IN);
						if (placeHolder.trim().equals("")) {
							if (paramName.startsWith("[") && paramName.endsWith("]")) {
								placeHolder = new String(paramName.substring(1, paramName.length() - 1));								
							} else {
								placeHolder = paramName;
							}
						}
						if (caseInsensitive) {
							dynamicWhereBuilder.append("{fn ucase(");
						}
						dynamicWhereBuilder.append(paramName);
						if (caseInsensitive) {
							dynamicWhereBuilder.append(")}");
						}
						dynamicWhereBuilder.append(" " + comparison.sql());
						if (isComparisonIn) {
							dynamicWhereBuilder.append("(");
						} else {
							dynamicWhereBuilder.append(" ");
						}
						if (caseInsensitive) {
							dynamicWhereBuilder.append("{fn ucase(");
						}
						dynamicWhereBuilder.append(":" + placeHolder);
						if (caseInsensitive) {
							dynamicWhereBuilder.append(")}");
						}
						if (isComparisonIn) {
							dynamicWhereBuilder.append(")");
						}
						if (log.isDebugEnabled()) {
							log.debug("identified parameter: {[" + placeHolder + "] = [" + args[i] + "]}");
						}
						params.put(placeHolder, args[i]);
					}
				}
			}
			while (openParens > 0) {
				dynamicWhereBuilder.append(")");
				openParens--;
			}
			
			StringBuilder finalBuilder = new StringBuilder();
			if (staticWhereBuilder.length() > 0) {
				finalBuilder.append(staticWhereBuilder.toString());
				if (dynamicWhereBuilder.length() > 0) {
					finalBuilder.append(" and (").append(dynamicWhereBuilder.toString()).append(")");
				}
			} else {
				finalBuilder.append(dynamicWhereBuilder.toString());
			}
			sb.append(finalBuilder.toString());
		}

		private DTOClass getDTOClass(Method method) {
			DTOClass type = method.getAnnotation(DTOClass.class);
			if (type == null) {
				type = getDTOClass(proxyClazz);
			}
			return type;
		}
		
		private DTOClass getDTOClass(Class<?> clazz) {
			DTOClass type = clazz.getAnnotation(DTOClass.class);
			if (type == null) {
				Class<?>[] interfaces = clazz.getInterfaces();
				for (int i = 0; i < interfaces.length; i++) {
					type = getDTOClass(interfaces[i]);
					if (type != null) {
						return type;
					}
				}
				Class<?> superClass = clazz.getSuperclass();
				if (superClass != null) {
					type = getDTOClass(superClass);	
				}
				
			}
			return type;
		}

		
	}
		
}
